<?php

if ( ! defined( 'ABSPATH' ) ) exit;

class BT_Tools_Settings {

	/**
	 * The single instance of BT_Tools_Settings.
	 * @var 	object
	 * @access  private
	 * @since 	2.0.0
	 */
	private static $_instance = null;

	/**
	 * The main plugin object.
	 * @var 	object
	 * @access  public
	 * @since 	2.0.0
	 */
	public $parent = null;

	/**
	 * Prefix for plugin settings.
	 * @var     string
	 * @access  public
	 * @since   2.0.0
	 */
	public $base = '';

	/**
	 * Available settings for plugin.
	 * @var     array
	 * @access  public
	 * @since   2.0.0
	 */
	public $settings = array();

	/**
	 * An array of Custom Post Type data that is currently being processed (e.g. when setting up CPTs from VenYou rtype data).
	 * @var     array
	 * @access  public
	 * @since   2.0.1
	 */
	public $cpt = array();

	/**
	 * The meta field slug base that is currently being processed.
	 * @var     string
	 * @access  public
	 * @since   2.0.1
	 */
	public $cpt_meta_slug_base = '';

	/**
     * Server path to where we place imported files
	 * @var     string
	 * @access  public
	 * @since   2.0.1
	 */
    public $import_file_path = '';

	/**
	 * URL to where we place imported files
	 * @var     string
	 * @access  public
	 * @since   2.0.1
	 */
    public $import_file_url = '';

	/**
	 * The first available ID beyond all existing Resources being imported.
     * This helps us force any attachments or other new posts to start at IDs higher than what's in VenYou, avoiding collision.
	 * @var     int
	 * @access  public
	 * @since   2.0.1
	 */
    public $first_available_id = 0;

	public function __construct ( $parent ) {
		$this->parent = $parent;

		$this->base = 'bt_';

		// prep import file path and url for uploads we bring over
		$uploads = wp_upload_dir();
		$this->import_file_path = $uploads['basedir'] . '/vy/';
		$this->import_file_url = $uploads['baseurl'] . '/vy/';

		// Initialise settings
		add_action( 'init', array( $this, 'init_settings' ), 11 );

		// Register plugin settings
		add_action( 'admin_init' , array( $this, 'register_settings' ) );

		// Add settings page to menu
		add_action( 'admin_menu' , array( $this, 'add_menu_item' ) );

		// Add settings link to plugins page
		add_filter( 'plugin_action_links_' . plugin_basename( $this->parent->file ) , array( $this, 'add_settings_link' ) );

		// Add ajax hook for VY Import
		add_action( 'wp_ajax_bt_vy_import' , array( $this, 'bt_vy_import' ) );
	}

	/**
	 * Initialise settings
	 * @return void
	 */
	public function init_settings () {
		$this->settings = $this->settings_fields();
	}

	/**
	 * Add settings page to admin menu
	 * @return void
	 */
	public function add_menu_item () {
		$page = add_options_page( __( 'BT Tools', 'bt-tools' ) , __( 'BT Tools', 'bt-tools' ) , 'manage_options' , $this->parent->_token . '_settings' ,  array( $this, 'settings_page' ) );
		add_action( 'admin_print_styles-' . $page, array( $this, 'settings_assets' ) );
	}

	/**
	 * Load settings JS & CSS
	 * @return void
	 */
	public function settings_assets () {

		// We're including the farbtastic script & styles here because they're needed for the colour picker
		// If you're not including a colour picker field then you can leave these calls out as well as the farbtastic dependency for the wpt-admin-js script below
		wp_enqueue_style( 'farbtastic' );
    	wp_enqueue_script( 'farbtastic' );

    	// We're including the WP media scripts here because they're needed for the image upload field
    	// If you're not including an image upload then you can leave this function call out
    	wp_enqueue_media();

    	wp_register_script( $this->parent->_token . '-settings-js', $this->parent->assets_url . 'js/settings' . $this->parent->script_suffix . '.js', array( 'farbtastic', 'jquery' ), '2.0.0' );
    	wp_enqueue_script( $this->parent->_token . '-settings-js' );
	}

	/**
	 * Add settings link to plugin list table
	 * @param  array $links Existing links
	 * @return array 		Modified links
	 */
	public function add_settings_link ( $links ) {
		$settings_link = '<a href="options-general.php?page=' . $this->parent->_token . '_settings">' . __( 'Settings', 'bt-tools' ) . '</a>';
  		array_push( $links, $settings_link );
  		return $links;
	}

	/**
	 * Build settings fields
	 * @return array Fields to be displayed on settings page
	 */
	private function settings_fields () {

		$settings['site_options'] = array(
			'title'					=> __( 'Site Options', 'bt-tools' ),
			'description'			=> __( 'Basic settings for your website.', 'bt-tools' ),
			'fields'				=> array(
				array(
					'id' 			=> 'admin_logo',
					'label'			=> __( 'Admin Logo' , 'bt-tools' ),
					'description'	=> __( 'Choose an image to display on the admin login page, instead of the WordPress Logo.', 'bt-tools' ),
					'type'			=> 'image',
					'default'		=> '',
					'placeholder'	=> ''
				),
				array(
					'id' 			=> 'admin_color',
					'label'			=> __( 'Admin Accent Color', 'bt-tools' ),
					'description'	=> __( 'Choose a brand color to accent the admin tools.', 'bt-tools' ),
					'type'			=> 'color',
					'default'		=> '#3fbced'
				),
			),
		);

		$settings['fonts'] = array(
			'title'					=> __( 'Fonts', 'bt-tools' ),
			'description'			=> __( 'Enable and set web fonts.', 'bt-tools' ),
			'fields'				=> array(
				array(
					'id' 			=> 'fonts_on',
					'label'			=> __( 'Enable Web Fonts', 'bt-tools' ),
					'description'	=> __( 'This enables fonts in general on the front end. Use fonts from one or more providers below as needed.', 'bt-tools' ),
					'type'			=> 'checkbox',
					'default'		=> ''
				),
				array(
					'id' 			=> 'fonts_loader_version',
					'label'			=> __( 'Web Font Loader Version' , 'bt-tools' ),
					'description'	=> __( '<em><a href="https://developers.google.com/speed/libraries/#web-font-loader" target="_blank">Available versions</a>. Defaults to ' . BT_FONT_LOADER_DEFAULT_VERSION . '.</em>', 'bt-tools' ),
					'type'			=> 'text',
					'size'          => 6,
					'default'		=> BT_FONT_LOADER_DEFAULT_VERSION,
					'placeholder'	=> __( BT_FONT_LOADER_DEFAULT_VERSION, 'bt-tools' )
				),
				array(
					'id' 			=> 'fonts_google',
					'label'			=> __( 'Google Fonts Families' , 'bt-tools' ),
					'description'	=> __( 'Enter one family declaration per line.', 'bt-tools' ),
					'type'			=> 'textarea',
					'default'		=> '',
					'placeholder'	=> __( 'Open Sans Condensed:300,700' . PHP_EOL . 'Droid Sans' . PHP_EOL . 'Droid Serif:bold', 'bt-tools' )
				),
				array(
					'id' 			=> 'fonts_fontscom_project_id',
					'label'			=> __( 'Fonts.com project ID' , 'bt-tools' ),
					'description'	=> '',
					'type'			=> 'text',
					'size'          => 40,
					'default'		=> '',
					'placeholder'	=> __( 'a1172442-f05d-4069-8cc2-65f9aba1406b', 'bt-tools' )
				),
			),
		);

		$settings['social_links'] = array(
			'title'					=> __( 'Social Links', 'bt-tools' ),
			'description'			=> __( 'Define the social links that appear on the website.', 'bt-tools' ),
			'fields'				=> array(),
		);
		$social_options = array(
			''              => '',
			'blog'          => 'Blog',
			'enews'         => 'E-Newsletter Signup',
			'facebook'      => 'Facebook',
			'googleplus'    => 'Google+',
			'instagram'     => 'Instagram',
			'linkedin'      => 'LinkedIn',
			'pinterest'     => 'Pinterest',
			'twitter'       => 'Twitter',
			'vimeo'         => 'Vimeo',
			'youtube'       => 'YouTube',
			'other'         => 'Other',
		);
		for ( $i = 1; $i <= BT_SOCIAL_LINK_SLOTS; $i++ ) {
			$settings['social_links']['fields'][] = array(
				'id' 			=> 'social_link_' . $i . '_section',
				'label'			=> __( '<h3>LINK ' . $i . '</h3>' , 'bt-tools' ),
				'type'			=> 'blank',
			);
			$settings['social_links']['fields'][] = array(
				'id' 			=> 'social_link_' . $i . '_type',
				'label'			=> __( 'Type' , 'bt-tools' ),
				'description'	=> '',
				'type'			=> 'select',
				'options'		=> $social_options,
				'default'		=> ''
			);
			$settings['social_links']['fields'][] = array(
				'id' 			=> 'social_link_' . $i . '_text',
				'label'			=> __( 'Display Text' , 'bt-tools' ),
				'description'	=> '',
				'type'			=> 'text',
				'default'		=> '',
				'placeholder'	=> ''
			);
			$settings['social_links']['fields'][] = array(
				'id' 			=> 'social_link_' . $i . '_url',
				'label'			=> __( 'URL' , 'bt-tools' ),
				'description'	=> '',
				'type'			=> 'url',
				'size'          => 50,
				'default'		=> '',
				'placeholder'	=> ''
			);
		}

		$settings['vy_import'] = array(
			'title'					=> __( 'VenYou Import', 'bt-tools' ),
			'description'			=> __( 'Pull data from VenYou into this website, and create custom content types from it.' . (get_option( $this->base . 'vy_site_url' ) ? '' : ' You must first save a site URL to import from, then you will have options to choose resource types and finally run the import.'), 'bt-tools' ),
			'fields'				=> array(
				array(
					'id' 			=> 'vy_site_url',
					'label'			=> __( 'URL of Site to Import' , 'bt-tools' ),
					'description'	=> 'Enter the base URL of the site to pull data from.',
					'type'			=> 'url',
					'size'          => 50,
					'default'		=> '',
					'placeholder'	=> __( 'http://w3.bfo.jeremy.dev/', 'bt-tools' )
				),
				array(
					'id' 			=> 'vy_rtypes_to_import',
					'label'			=> __( 'Resource Types to Import', 'bt-tools' ),
					'type'			=> 'hidden',
					'default'		=> ''
				),
				array(
					'id' 			=> 'vy_import_run',
					'label'			=> __( 'Run Import', 'bt-tools' ),
					'description'	=> __( 'Once you have entered a site to import from, and selected which resource types to import, checking this box will run the import itself upon clicking "Save Settings."', 'bt-tools' ),
					'type'			=> 'checkbox',
					'default'		=> ''
				),
			),
		);

		/*
		$settings['standard'] = array(
			'title'					=> __( 'Standard', 'bt-tools' ),
			'description'			=> __( 'These are fairly standard form input fields.', 'bt-tools' ),
			'fields'				=> array(
				array(
					'id' 			=> 'text_field',
					'label'			=> __( 'Some Text' , 'bt-tools' ),
					'description'	=> __( 'This is a standard text field.', 'bt-tools' ),
					'type'			=> 'text',
					'default'		=> '',
					'placeholder'	=> __( 'Placeholder text', 'bt-tools' )
				),
				array(
					'id' 			=> 'password_field',
					'label'			=> __( 'A Password' , 'bt-tools' ),
					'description'	=> __( 'This is a standard password field.', 'bt-tools' ),
					'type'			=> 'password',
					'default'		=> '',
					'placeholder'	=> __( 'Placeholder text', 'bt-tools' )
				),
				array(
					'id' 			=> 'secret_text_field',
					'label'			=> __( 'Some Secret Text' , 'bt-tools' ),
					'description'	=> __( 'This is a secret text field - any data saved here will not be displayed after the page has reloaded, but it will be saved.', 'bt-tools' ),
					'type'			=> 'text_secret',
					'default'		=> '',
					'placeholder'	=> __( 'Placeholder text', 'bt-tools' )
				),
				array(
					'id' 			=> 'text_block',
					'label'			=> __( 'A Text Block' , 'bt-tools' ),
					'description'	=> __( 'This is a standard text area.', 'bt-tools' ),
					'type'			=> 'textarea',
					'default'		=> '',
					'placeholder'	=> __( 'Placeholder text for this textarea', 'bt-tools' )
				),
				array(
					'id' 			=> 'single_checkbox',
					'label'			=> __( 'An Option', 'bt-tools' ),
					'description'	=> __( 'A standard checkbox - if you save this option as checked then it will store the option as \'on\', otherwise it will be an empty string.', 'bt-tools' ),
					'type'			=> 'checkbox',
					'default'		=> ''
				),
				array(
					'id' 			=> 'select_box',
					'label'			=> __( 'A Select Box', 'bt-tools' ),
					'description'	=> __( 'A standard select box.', 'bt-tools' ),
					'type'			=> 'select',
					'options'		=> array( 'drupal' => 'Drupal', 'joomla' => 'Joomla', 'wordpress' => 'WordPress' ),
					'default'		=> 'wordpress'
				),
				array(
					'id' 			=> 'radio_buttons',
					'label'			=> __( 'Some Options', 'bt-tools' ),
					'description'	=> __( 'A standard set of radio buttons.', 'bt-tools' ),
					'type'			=> 'radio',
					'options'		=> array( 'superman' => 'Superman', 'batman' => 'Batman', 'ironman' => 'Iron Man' ),
					'default'		=> 'batman'
				),
				array(
					'id' 			=> 'multiple_checkboxes',
					'label'			=> __( 'Some Items', 'bt-tools' ),
					'description'	=> __( 'You can select multiple items and they will be stored as an array.', 'bt-tools' ),
					'type'			=> 'checkbox_multi',
					'options'		=> array( 'square' => 'Square', 'circle' => 'Circle', 'rectangle' => 'Rectangle', 'triangle' => 'Triangle' ),
					'default'		=> array( 'circle', 'triangle' )
				)
			)
		);

		$settings['extra'] = array(
			'title'					=> __( 'Extra', 'bt-tools' ),
			'description'			=> __( 'These are some extra input fields that maybe aren\'t as common as the others.', 'bt-tools' ),
			'fields'				=> array(
				array(
					'id' 			=> 'number_field',
					'label'			=> __( 'A Number' , 'bt-tools' ),
					'description'	=> __( 'This is a standard number field - if this field contains anything other than numbers then the form will not be submitted.', 'bt-tools' ),
					'type'			=> 'number',
					'default'		=> '',
					'placeholder'	=> __( '42', 'bt-tools' )
				),
				array(
					'id' 			=> 'colour_picker',
					'label'			=> __( 'Pick a colour', 'bt-tools' ),
					'description'	=> __( 'This uses WordPress\' built-in colour picker - the option is stored as the colour\'s hex code.', 'bt-tools' ),
					'type'			=> 'color',
					'default'		=> '#21759B'
				),
				array(
					'id' 			=> 'an_image',
					'label'			=> __( 'An Image' , 'bt-tools' ),
					'description'	=> __( 'This will upload an image to your media library and store the attachment ID in the option field. Once you have uploaded an imge the thumbnail will display above these buttons.', 'bt-tools' ),
					'type'			=> 'image',
					'default'		=> '',
					'placeholder'	=> ''
				),
				array(
					'id' 			=> 'multi_select_box',
					'label'			=> __( 'A Multi-Select Box', 'bt-tools' ),
					'description'	=> __( 'A standard multi-select box - the saved data is stored as an array.', 'bt-tools' ),
					'type'			=> 'select_multi',
					'options'		=> array( 'linux' => 'Linux', 'mac' => 'Mac', 'windows' => 'Windows' ),
					'default'		=> array( 'linux' )
				)
			)
		);
		*/

		$settings = apply_filters( $this->parent->_token . '_settings_fields', $settings );

		return $settings;
	}

	/**
	 * Register plugin settings
	 * @return void
	 */
	public function register_settings () {
		if ( is_array( $this->settings ) ) {

			// Check posted/selected tab
			$current_section = '';
			if ( isset( $_POST['tab'] ) && $_POST['tab'] ) {
				$current_section = $_POST['tab'];
			} else {
				if ( isset( $_GET['tab'] ) && $_GET['tab'] ) {
					$current_section = $_GET['tab'];
				}
			}

			foreach ( $this->settings as $section => $data ) {

				if ( $current_section && $current_section != $section ) continue;

				// Add section to page
				add_settings_section( $section, $data['title'], array( $this, 'settings_section' ), $this->parent->_token . '_settings' );

				foreach ( $data['fields'] as $field ) {

					// Validation callback for field
					$validation = '';
					if ( isset( $field['callback'] ) ) {
						$validation = $field['callback'];
					}

					// Register field
					$option_name = $this->base . $field['id'];
					register_setting( $this->parent->_token . '_settings', $option_name, $validation );

					// hide the import details fields if there is no VenYou site url set to pull from
					$show_field = true;
					if ( in_array( $field['id'], array( 'vy_rtypes_to_import', 'vy_import_run' ) ) && !get_option( $this->base . 'vy_site_url' ) ) {
						$show_field = false;
					}

					// Add field to page
					if ( $show_field ) {
						add_settings_field( $field['id'], $field['label'], array( $this->parent->admin, 'display_field' ), $this->parent->_token . '_settings', $section, array( 'field' => $field, 'prefix' => $this->base ) );
					}
				}

				if ( ! $current_section ) break;
			}
		}
	}

	public function settings_section ( $section ) {
		$html = '<p> ' . $this->settings[ $section['id'] ]['description'] . '</p>' . "\n";
		echo $html;
	}

	/**
	 * Load settings page content
	 * @return void
	 */
	public function settings_page () {

		// Build page HTML
		$html = '<div class="wrap" id="' . $this->parent->_token . '_settings">' . "\n";
			$html .= '<h2>' . __( 'BT Tools' , 'bt-tools' ) . '</h2>' . "\n";

			$tab = '';
			if ( isset( $_GET['tab'] ) && $_GET['tab'] ) {
				$tab .= $_GET['tab'];
			}

			// Show page tabs
			if ( is_array( $this->settings ) && 1 < count( $this->settings ) ) {

				$html .= '<h2 class="nav-tab-wrapper">' . "\n";

				$c = 0;
				foreach ( $this->settings as $section => $data ) {

					// Set tab class
					$class = 'nav-tab';
					if ( ! isset( $_GET['tab'] ) ) {
						if ( 0 == $c ) {
							$class .= ' nav-tab-active';
						}
					} else {
						if ( isset( $_GET['tab'] ) && $section == $_GET['tab'] ) {
							$class .= ' nav-tab-active';
						}
					}

					// Set tab link
					$tab_link = add_query_arg( array( 'tab' => $section ) );
					if ( isset( $_GET['settings-updated'] ) ) {
						$tab_link = remove_query_arg( 'settings-updated', $tab_link );
					}

					// Output tab
					$html .= '<a href="' . $tab_link . '" class="' . esc_attr( $class ) . '">' . esc_html( $data['title'] ) . '</a>' . "\n";

					++$c;
				}

				$html .= '</h2>' . "\n";
			}

			$html .= '<form method="post" action="options.php" enctype="multipart/form-data">' . "\n";

				// Get settings fields
				ob_start();
				settings_fields( $this->parent->_token . '_settings' );
				do_settings_sections( $this->parent->_token . '_settings' );
				$html .= ob_get_clean();

				$html .= '<p class="submit">' . "\n";
					$html .= '<input type="hidden" name="tab" value="' . esc_attr( $tab ) . '" />' . "\n";
					$html .= '<input name="Submit" type="submit" class="button-primary" value="' . esc_attr( __( 'Save Settings' , 'bt-tools' ) ) . '" />' . "\n";
				$html .= '</p>' . "\n";
			$html .= '</form>' . "\n";
		$html .= '</div>' . "\n";

		echo $html;
	}

	/**
	 * Main BT_Tools_Settings Instance
	 *
	 * Ensures only one instance of BT_Tools_Settings is loaded or can be loaded.
	 *
	 * @since 2.0.0
	 * @static
	 * @see BT_Tools()
	 * @return Main BT_Tools_Settings instance
	 */
	public static function instance ( $parent ) {
		if ( is_null( self::$_instance ) ) {
			self::$_instance = new self( $parent );
		}
		return self::$_instance;
	} // End instance()

	/**
	 * Cloning is forbidden.
	 *
	 * @since 2.0.0
	 */
	public function __clone () {
		_doing_it_wrong( __FUNCTION__, __( 'Cheatin&#8217; huh?' ), $this->parent->_version );
	} // End __clone()

	/**
	 * Unserializing instances of this class is forbidden.
	 *
	 * @since 2.0.0
	 */
	public function __wakeup () {
		_doing_it_wrong( __FUNCTION__, __( 'Cheatin&#8217; huh?' ), $this->parent->_version );
	} // End __wakeup()

	/**
	 * Do AJAX handling for VY Import
	 *
	 * @since 2.0.1
     * @public
     * @return void
	 */
	public function bt_vy_import() {

        // set ini var for memory limit and max execution time to allow for processing lots of data
		ini_set('memory_limit', '1536M');
		ini_set('max_execution_time', 60*30);

		// get site url to import from and validate it
		$vy_url = filter_var( get_option( $this->base . 'vy_site_url' ), FILTER_VALIDATE_URL );
		if ( !$vy_url ) {
			echo 'Invalid URL to import from!';
			wp_die();
		}

		// get rtypes that are selected for import, and their custom post type data (it's all together in json, with rtype ID as key)
		$cpt_data_wp = $this->parent->vy_get_rtypes( true );
		// loop to get rtype IDs that are to be imported
		$rtypes_to_import = array();
		foreach ( $cpt_data_wp as $rtype_id => $rtype_json ) {
			$rtypes_to_import[] = $rtype_id;
        }

		// check if we're going do the import (get resource data with the request and process it)
		$run_import = get_option( $this->base . 'vy_import_run' );

		// check when the last import was
		$last_import = get_option( $this->base . 'vy_last_import' );
        if ( !$last_import ) {
            $last_import = 'never';
        }
		?>
        <div class="rtype">
            Last import: <span id="last_import_time"><?php echo $last_import; ?></span>
        </div>
		<?php

		// make request to get json
		$json_uri = rtrim( $vy_url, '/' ) . '/utilities/resources.php';
		if ( $run_import ) {
		    foreach ( $rtypes_to_import as $rtype_id ) {
			    $json_uri = add_query_arg( 'rtypes_to_import[]', $rtype_id, $json_uri);
            }
			?>
            <div class="rtype vy_success">
                Import ran, using <a href="<?php echo $json_uri; ?>" target="_blank">this data set</a> (<a href="<?php echo add_query_arg( 'pretty', 1, $json_uri ); ?>" target="_blank">pretty</a>).
            </div>
			<?php
		}

		$api = wp_remote_get(
			$json_uri,
			array(
				'timeout' => 1800,
				'httpversion' => '1.1'
			)
		);

		if ( is_a( $api, 'WP_Error' ) ) {
			echo 'API call resulted in error!<br>URI: ' . $json_uri . '<br>ERROR MESSAGE: <pre>' . $api->get_error_message() . '</pre>';
			wp_die();

		} else if ( $api['response']['code'] == '200' ) {

		    // read json results from response body
			$api_data = json_decode( $api['body'] );
			//echo '<pre>' . print_r( $api_data, true ) . '</pre>';

            // make a path to where we'll copy files into if we import
			wp_mkdir_p( $this->import_file_path );

			// loop resource types from json to create a map of slugs by ID, for lookup as we fully process each rtype
            $rtype_slug_map = array();
            $general_assoc_rtypes = array();
			foreach ( $api_data as $slug => $rtype ) {
				$rtype_slug_map[$rtype->intID] = $slug;
				if ( $rtype->intAssocAsGeneralResource && in_array( $rtype->intID, $rtypes_to_import ) ) {
					$general_assoc_rtypes[] = $slug; // no ID for key so that ACF sorts slugs alphabetically (it was sorting by ID when it existed)
                }
			}

            // loop resource types from json to process each rtype
            $count = 0;
			foreach ( $api_data as $slug => $rtype ) {
			    $count++;

			    // get the max resource id from the data (which is the same in every rtype as it just has to be in them)
                // and set the first available ID WP can use when adding images (so as to avoid using IDs of yet-to-be-imported Resources)
                if ( $this->first_available_id === 0 ) {
                    // first available id has NOT been set yet at all; init it
	                $this->first_available_id = 1 + $rtype->max_resource_id;
                }

                // get a plural slug
                $slug_plural = sanitize_title( $rtype->strNamePlural );

			    // build data to save in options for custom post type setup
                $this->cpt = array(
                    'slug' => $slug,
                    'single' => __( $rtype->strName, 'bt-tools' ),
	                'plural' => __( $rtype->strNamePlural, 'bt-tools' ),
	                'description' => '',
	                'options' => array(
		                /**
		                 * outcommented vars are defaults we may want to override for some complex rtypes
                         * @see https://developer.wordpress.org/reference/functions/register_post_type/
                         * @see http://justintadlock.com/archives/2013/09/13/register-post-type-cheat-sheet
		                 */
		                //'public'                => true,
		                //'publicly_queryable'    => true,
		                //'exclude_from_search'   => false,
		                //'show_ui'               => true,
		                //'show_in_menu'          => true, // could make something with submenus, e.g. "Resources" which has all VY CPTs
		                //'show_in_nav_menus'     => true,
		                //'can_export'            => true,
		                'rewrite'               => array(
                            'slug' => $slug_plural,
                            'with_front' => false,
                        ),
		                // 'capability_type'       => array( $rtype->strName, $rtype->strNamePlural ), // could create more simple caps, see cheatsheet
		                //'has_archive'           => true,
		                'hierarchical'          => false, // default in WPT is true
		                'supports'              => array(
		                    'title',
                            'editor',
                            //'excerpt',
                            'thumbnail',
			                'author',
			                'revisions',
                            /*
                            'comments',
                            'trackbacks',
                            'custom-fields',
                            'page-attributes', // parent, sort order
                            'post-formats',
                             */
                        ),
		                'menu_position'         => 27 + $count, // '26.' . str_pad( $count, 2, '0', STR_PAD_LEFT ), // comments is 25, next default (separator) is 59
		                'menu_icon'             => 'dashicons-admin-post',
		                'delete_with_user'      => false,
                        'taxonomies'            => array(), // standard arg for register_post_type() which would be like ['partner_categories']
                        'taxonomies_extended'   => array(), // each key is tax_slug with value that has args for register_taxonomy() like single, plural
                    ),
                    'meta' => array(),
                );

				// set up fields for which we allow admins to update data
                // use "-VY-" between slugs that are more than one level deep in the cpt array
				$editable_fields = array(
					'slug' => array(
						'label' => 'Slug',
						'value' => $slug,
					),
					'options-VY-rewrite-VY-slug' => array(
						'label' => 'URL/Rewrite Slug',
						'value' => $this->cpt['options']['rewrite']['slug'],
					),
					'single' => array(
						'label' => 'Single',
						'value' => $this->cpt['single'],
					),
					'plural' => array(
						'label' => 'Plural',
						'value' => $this->cpt['plural'],
					),
                    'options-VY-menu_icon' => array(
	                    'label' => 'Menu Icon',
	                    'value' => $this->cpt['options']['menu_icon'],
                    ),
				);

				// check for any explicitly unsupported rtype config
                if ( $rtype->boolUseAsWebsiteUser ) {
	                wp_die( 'Fatal Error: this tool cannot yet handle resources used as website users.' );
                }
				if ( $rtype->intUseSort ) {
					wp_die( 'Fatal Error: this tool cannot yet handle resources with user-defined sort orders.' ); // not in modern rtype database table anyway!
				}
				if ( $rtype->intUseAsStoreItems ) {
					wp_die( 'Fatal Error: this tool cannot yet handle resources as store items.' );
				}
				if ( $rtype->intAssocPTreePages ) {
					wp_die( 'Fatal Error: this tool cannot yet handle page tree pages associated to resources.' );
				}
				if ( $rtype->intLocationForGMap ) {
					wp_die( 'Fatal Error: this tool cannot yet handle resources with Location used for GMaps.' ); // we could probably use a Google Map field type instead of Text for Location, yes?
				}

				// disable thumbnail if rtype doesn't use an image
				if ( !$rtype->intUseImage ) {
					$this->bt_vy_import_rtype_remove_supports( 'thumbnail' );
				}

				// use excerpt if rtype has overview and desc field
                if ( $rtype->intUseOverview && $rtype->intUseDesc ) {
	                $this->cpt['options']['supports'][] = 'excerpt';
                }

				// set any rtype data that has been changed on the WP side (e.g. rtype name or slug)
				$cpt_wp = $cpt_data_wp[$rtype->intID];
                if ( $cpt_wp['slug'] ) {
                    // there is data in the options table for this rtype, sp we can safely use it (otherwise we'd clobber venyou data on the first time)
	                foreach ( $editable_fields as $field_slug => $field_data ) {
	                    // check for depth of field slug (e.g. if part of options data)
                        $field_slug_parts = explode( '-VY-', $field_slug );
		                // set both the CPT value (json data) and field value (form input corresponding to json data) to what was in the options table
		                if ( count($field_slug_parts) == 1 ) {
			                $this->cpt[$field_slug] = $editable_fields[$field_slug]['value'] = $cpt_wp[$field_slug];
		                } else if ( count($field_slug_parts) == 2 ) {
			                $this->cpt[$field_slug_parts[0]][$field_slug_parts[1]] = $editable_fields[$field_slug]['value'] = $cpt_wp[$field_slug_parts[0]][$field_slug_parts[1]];
		                } else if ( count($field_slug_parts) == 3 ) {
			                $this->cpt[$field_slug_parts[0]][$field_slug_parts[1]][$field_slug_parts[2]] = $editable_fields[$field_slug]['value'] = $cpt_wp[$field_slug_parts[0]][$field_slug_parts[1]][$field_slug_parts[2]];
		                } else {
		                    wp_die( 'Invalid number of parts to field_slug "' . $field_slug . '".' );
                        }
	                }
	                // use this slug as the post type slug
                    $post_type_slug = $cpt_wp['slug'];
                } else {
                    // use VenYou slug as the post type slug
	                $post_type_slug = $slug;
                }

				/** ****************************************************************************************************
				 * META -- build up based on rtype config – data elements of each meta field match up to ACF Pro fields
				 **************************************************************************************************** */
				$this->cpt_meta_slug_base = $this->base . $post_type_slug . '_';

				/* Standard Meta Fields, grouped below details in editor */

				// download/file
				if ( $this->bt_vy_import_rtype_has_file( $slug, $rtype->intUseDownload ) ) {
					$this->bt_vy_import_rtype_add_field_meta(
						'File', // do not change field title or use VY label - code below depends on field key being "file"
						'file',
						array(
							'required' => 1,
							'vy_map_to' => 'strDownloadFilename',
						)
					);
				}

				// date range
				if ( $rtype->intUseDateRange ) {
					// single/start date
					$this->bt_vy_import_rtype_add_field_meta(
						'Single or Start Date',
						'date_picker',
						array(
							'required' => 1,
							'display_format' => 'F j, Y',
							'return_format' => 'F j, Y',
							'first_day' => 0,
							'vy_map_to' => 'strDate',
						)
					);
					// end date
					$this->bt_vy_import_rtype_add_field_meta(
						'End Date',
						'date_picker',
						array(
							'display_format' => 'F j, Y',
							'return_format' => 'F j, Y',
							'first_day' => 0,
							'vy_map_to' => 'strEndDate',
						)
					);
				}

				// time
				if ( $rtype->intUseTime) {
					// start time
					$this->bt_vy_import_rtype_add_field_meta(
						'Start Time',
						'time_picker',
						array(
							'display_format' => 'g:i a',
							'return_format' => 'g:i a',
							'vy_map_to' => 'strStartTime',
						)
					);
					// end time
					$this->bt_vy_import_rtype_add_field_meta(
						'End Time',
						'time_picker',
						array(
							'display_format' => 'g:i a',
							'return_format' => 'g:i a',
							'vy_map_to' => 'strEndTime',
						)
					);
				}

				// price
				if ( $rtype->intUsePrice && ! $rtype->intUseAsStoreItems ) {
				    switch ( $rtype->strName ) {
					    case 'Event':
						    $instructions = 'Leave price blank to omit from display. Set price to zero to display as "free."';
						    break;
					    case 'Partner': // BFO
						    $instructions = '';
						    break;
					    default:
						    $instructions = '';
						    break;
                    }
					$this->bt_vy_import_rtype_add_field_meta(
						( $rtype->strPriceLabel ? $rtype->strPriceLabel : 'Price' ),
						'number',
						array(
							'required' => ($rtype->intPriceRequired ? 1 : 0 ),
							'instructions' => $instructions,
                            'min' => 0,
							'prepend' => '$',
							'vy_map_to' => 'floatPrice',
						)
					);
				}

				// source
                if ( $rtype->intUseSource ) {
	                switch ( $rtype->strName ) {
		                case 'Partner': // BFO
			                $type = 'number';
		                    $maxlength = '';
		                    $placeholder = '';
			                break;
		                default:
			                $type = 'text';
			                $maxlength = '100';
			                $placeholder = __( 'Sample ' . $rtype->strSourceLabel, 'bt-tools' );
			                break;
	                }
	                $this->bt_vy_import_rtype_add_field_meta(
                        $rtype->strSourceLabel,
		                $type,
                        array(
	                        'placeholder' => $placeholder,
	                        'maxlength' => $maxlength,
                            'vy_map_to' => 'strSource',
                        )
                    );
	                // source url
                    if ( $rtype->intUseSourceURL ) {
	                    $this->bt_vy_import_rtype_add_field_meta(
		                    $rtype->strSourceLabel . ' URL',
                            'url',
                            array(
	                            'vy_map_to' => 'strSourceURL',
                            )
	                    );
                    }
                }

				// location
				if ( $rtype->intUseLocation ) {
					$this->bt_vy_import_rtype_add_field_meta(
						'Location',
						'text',
						array(
							'required' => ($rtype->intLocationRequired ? 1 : 0 ),
							'placeholder' => __( 'Sample Location Name', 'bt-tools' ),
							'vy_map_to' => 'strLocation',
						)
					);
					// location url
					if ( $rtype->intUseLocationURL ) {
						$this->bt_vy_import_rtype_add_field_meta(
							'Location URL',
							'url',
							array(
								'vy_map_to' => 'strLocationURL',
							)
						);
					}
				}

				// link
				if ( $rtype->intUseURL ) {
					$this->bt_vy_import_rtype_add_field_meta(
						$rtype->strURLLabel  . ' URL',
						'url',
                        array(
	                        'vy_map_to' => 'strLinkURL',
                        )
					);
					// link text/label
					if ( $rtype->intURLUseLinkText ) {
						$this->bt_vy_import_rtype_add_field_meta(
							$rtype->strURLLabel . ' Text',
							'text',
							array(
								'placeholder' => __( 'Sample ' . $rtype->strURLLabel, 'bt-tools' ),
								'vy_map_to' => 'strLinkText',
							)
						);
                    }
					// link use new window option
					if ( $rtype->intURLUseNewWinOpt ) {
						$this->bt_vy_import_rtype_add_field_meta(
							'Open in New Tab',
							'true_false',
                            array(
                                'vy_map_to' => 'boolLinkNewWin',
                            )
						);
                    }

                    // if this is the standard "link" resource type, adjust other relevant bits for WP CPT
                    if ( $post_type_slug == 'link' ) {
					    // remove editor/details
                        $this->bt_vy_import_rtype_remove_supports( 'editor' );
                    }
				}

				// code
                if ( $rtype->intUseCode && $rtype->strCodeLabel != 'YouTube ID (Legacy)' ) {
	                $this->bt_vy_import_rtype_add_field_meta(
                        ( $rtype->strCodeLabel ? $rtype->strCodeLabel : 'Code' ),
		                'text',
		                array(
                            'required' => ($rtype->intCodeRequired ? 1 : 0 ),
			                'vy_map_to' => 'strCode',
                        )
	                );
                }

				// position
				if ( $rtype->intUsePosition ) {
					$this->bt_vy_import_rtype_add_field_meta(
						'Position',
						'text',
						array(
							'vy_map_to' => 'strPosition',
						)
					);
				}

				// email
				if ( $rtype->intUseEmail ) {
					$this->bt_vy_import_rtype_add_field_meta(
						'Email Address',
						'email',
						array(
							'required' => ($rtype->intEmailRequired ? 1 : 0 ),
							'vy_map_to' => 'strEmail',
						)
					);
				}

				// phone numbers
                if ( $rtype->intUsePhoneNumbers ) {
	                $phone_sub_field_slug_base = $this->cpt_meta_slug_base . 'phone-numbers_';
	                $this->bt_vy_import_rtype_add_field_meta(
		                'Phone Numbers',
		                'repeater',
		                array(
			                'min' => 0,
                            'max' => 0,
                            'layout' => 'table',
                            'button_label' => 'Add Phone Number',
                            'collapsed' => '',
			                'sub_fields' => array(
                                array(
                                    'placeholder' => 'Direct',
                                    'prepend' => '',
                                    'append' => '',
                                    'key' => $phone_sub_field_slug_base . 'label',
                                    'label' => 'Label',
                                    'name' => 'label',
                                    'type' => 'text',
                                    'required' => 1,
                                ),
                                array(
                                    'default_value' => '',
                                    'min' => 200,
                                    'max' => 999,
                                    'step' => '',
                                    'placeholder' => 612,
                                    'prepend' => '',
                                    'append' => '-',
                                    'key' => $phone_sub_field_slug_base . 'area',
                                    'label' => 'Area Code',
                                    'name' => 'area_code',
                                    'type' => 'number',
                                    'required' => 1,
                                ),
                                array(
                                    'default_value' => '',
                                    'min' => 100,
                                    'max' => 999,
                                    'step' => '',
                                    'placeholder' => 555,
                                    'prepend' => '',
                                    'append' => '-',
                                    'key' => $phone_sub_field_slug_base . 'exchange',
                                    'label' => 'Exchange',
                                    'name' => 'exchange',
                                    'type' => 'number',
                                    'required' => 1,
                                ),
                                array(
                                    'default_value' => '',
                                    'min' => '',
                                    'max' => '',
                                    'step' => '',
                                    'placeholder' => 1212,
                                    'prepend' => '',
                                    'append' => '',
                                    'key' => $phone_sub_field_slug_base . 'suffix',
                                    'label' => 'Suffix',
                                    'name' => 'suffix',
                                    'type' => 'number',
                                    'required' => 1,
                                ),
                                array(
                                    'default_value' => '',
                                    'min' => '',
                                    'max' => '',
                                    'step' => '',
                                    'placeholder' => '',
                                    'prepend' => 'x',
                                    'append' => '',
                                    'key' => $phone_sub_field_slug_base . 'ext',
                                    'label' => 'Extension',
                                    'name' => 'extension',
                                    'type' => 'number',
                                    'required' => 0,
                                )
                            ),
			                'vy_map_to' => 'arrPhoneNumbers',
		                )
	                );
                }

				// multiple images for a gallery
				if ( $rtype->intUseMultiplePhotos ) {
					$phone_sub_field_slug_base = $this->cpt_meta_slug_base . 'images_';
					$this->bt_vy_import_rtype_add_field_meta(
						'Images',
						'gallery',
						array(
							'library' => 'all',
                            'insert' => 'append',
							'vy_map_to' => 'arrGalleryImages',
						)
					);
				}

				// address / google map
				if ( $rtype->intUseStreetAddress ) {
					$this->bt_vy_import_rtype_add_field_meta(
						'Address',
						'google_map',
						array(
							'center_lat' => '47.2371658',
                            'center_lng' => '-93.5302142',
                            'zoom' => 10,
							'vy_map_to' => 'arrMapData',
						)
					);
				}

				// arbitrary attributes
                if ( $rtype->intUseAttributes ) {

				    foreach ( $rtype->arrTypeAttributes as $attribute_id => $attribute ) {
				        // skip any known attributes we don't want to carry forward. we're doing this by name so we don't have to introduce client IDs and check for their attribute IDs.
                        $attributes_to_exclude = array(
                            'Exclude from Main Site', // BFO: flag for child-site-only rtypes; goes away with single site
                        );
                        if ( in_array( $attribute->strName, $attributes_to_exclude ) ) {
                            continue;
                        }

				        // check for certain attribute types, determine vars based on them
                        $attribute_type = 'text';
                        $attribute_maxlength = '';
                        switch ( $attribute->strCode ) {
                            case 'string-s':
	                            $attribute_type = 'text';
	                            $attribute_maxlength = $attribute->intMaxLength;
                                break;
	                        case 'string-l':
		                        $attribute_type = 'textarea';
		                        break;
	                        case 'email':
		                        $attribute_type = 'email';
		                        break;
	                        case 'bool':
		                        $attribute_type = 'true_false';
		                        break;
                            default:
                                // see Resource::SetAttributeData() for available types (except for option types which are handled separately, all as arrays)
                                wp_die( __METHOD__ . ' -- Fatal Error: this tool cannot yet handle attribute type "' . $attribute->strCode . '"' );
                        }

				        // add a meta field for current attribute
					    $this->bt_vy_import_rtype_add_field_meta(
						    $attribute->strName,
						    $attribute_type,
						    array(
							    'instructions' => $attribute->strNotes,
                                'required' => ($attribute->required ? 1 : 0 ),
                                'maxlength' => $attribute_maxlength,
							    'vy_map_to' => 'arrAttributes->' . $attribute_id . '->value',
						    )
					    );
				    }

                }

                /* Fields Related To Publication Status, in a box to the side below default status/date box */

				// expiration date
				if ( $rtype->intUseExpiration ) {
					$this->bt_vy_import_rtype_add_field_meta(
						'Expiration Date',
						'date_picker',
						array(
                            'instructions' => 'If set, this content won\'t appear on the website on or after that date.',
                            'display_format' => 'm-d-Y',
							'return_format' => 'm-d-Y',
                            'first_day' => 0,
							'vy_map_to' => 'strExpirationDate',
						)
					);
				}

				/* Related Content, in its own box below standard meta fields */

				// specific associated rtypes
                if ( count( $rtype->arrAssocTypeIDs ) ) {
                    //error_log( __METHOD__ . ' -- Processing specific assoc rtypes: ' . print_r( $rtype->arrAssocTypeIDs, 1 ));
                    foreach ( $rtype->arrAssocTypeIDs as $assoc_rtype ) {
                        // check for a max, which is there after a colon
	                    $assoc_rtype_parts = explode( ':', $assoc_rtype );
	                    $assoc_rtype_id = preg_replace( '/\D/', '', $assoc_rtype_parts[0] ); // get only digits (sometimes has a letter flag like "d" for "disable sort")
                        $assoc_rtype_max = $assoc_rtype_parts[1] ? $assoc_rtype_parts[1] : 0;
                        $assoc_rtype_slug = $rtype_slug_map[$assoc_rtype_id];
	                    //error_log( __METHOD__ . ' -- Processing specific assoc rtypes for ' . $slug . ': Assoc ID = ' . $assoc_rtype_id . ' / Assoc Slug = ' . $assoc_rtype_slug . ' / Max = ' . $assoc_rtype_max );
	                    if ( $assoc_rtype_slug ) {
		                    $this->bt_vy_import_rtype_add_field_meta(
			                    'Related ' . ( $assoc_rtype_max == 1 ? $api_data->{$assoc_rtype_slug}->strName : $api_data->{$assoc_rtype_slug}->strNamePlural ),
			                    'relationship',
			                    array(
				                    'instructions' => ( $assoc_rtype_max ? 'Maximum allowed: ' . $assoc_rtype_max : '' ),
                                    'post_type' => array( $assoc_rtype_slug ),
                                    'max' => $assoc_rtype_max,
                                    'elements' => array( 'featured_image' ),
				                    'vy_map_to' => 'arrAssocResourcesType' . $assoc_rtype_id,
			                    )
		                    );
                        }
                    }
                }

                // general associated resources
                if ( $rtype->intAssocGeneralResources ) {
	                $this->bt_vy_import_rtype_add_field_meta(
		                'General Related Resources',
		                'relationship',
		                array(
			                'post_type' => $general_assoc_rtypes,
			                'elements' => array( 'featured_image' ),
			                'vy_map_to' => 'arrAssocResourcesGeneral',
		                )
	                );
                }

				/** ****************************************************************************************************
				 * TAXONOMY -- build up based on rtype config
				 **************************************************************************************************** */
				$tax_slug = '';
				$rtype_categories = array();
				if ( $rtype->intUseCategories ) {
				    // get all categories as an array (passing through json encoding to convert from stdClass)
					$rtype_categories = json_decode( json_encode( $rtype->arrCategories ), true );
					$tax_slug = $post_type_slug . '_categories';
					$this->cpt['options']['taxonomies'][] = $tax_slug;
                    $this->cpt['options']['taxonomies_extended'][$tax_slug] = array(
                        'single' => $this->cpt['single'] . ' Category',
                        'plural' => $this->cpt['single'] . ' Categories',
                        'args' => array(),
                    );

                    // import categories if we're running a full import
                    if ( $run_import && in_array( $rtype->intID, $rtypes_to_import ) ) {
	                    //error_log( __METHOD__ . ' -- Going to import CATEGORIES for rtype ' . $rtype->intID . '/' . $slug . '!' );
	                    foreach ( $rtype->arrCategories as $category_id => $category ) {

		                    // make a slug for the term if it doesn't exist
		                    if ( !isset( $category->slug ) || !$category->slug ) {
			                    $category->slug = sanitize_title( $category->strTitle );
		                    }

		                    // check if the term exists, insert if not
		                    $term = term_exists( $category->slug, $tax_slug );
		                    if ( !$term ) {
			                    $term = wp_insert_term(
				                    $category->strTitle,
				                    $tax_slug,
				                    array(
					                    'description' => '',
					                    'slug'        => $category->slug,
					                    //'parent'      => $parent_term_id,
				                    )
			                    );
			                    if ( is_a( $term, 'WP_Error' ) ) {
			                        error_log( __METHOD__ . ' -- Inserting Term "' . $category->strTitle . '" resulted in WP Error: ' . $term->get_error_message() );
                                }
		                    }

		                    // add the new term to our resources categories array so that we can use the data below when importing resources
		                    $rtype_categories[$category_id]['term'] = $term;

	                    }
                    }
				}

				/** ****************************************************************************************************
				 * RESOURCES -- import if requested!
				 **************************************************************************************************** */
				$current_rtype_imported = false;
				if ( $run_import && in_array( $rtype->intID, $rtypes_to_import ) ) {
                    //error_log( __METHOD__ . ' -- Going to import RESOURCES for rtype ' . $rtype->intID . '/' . $slug . '!' );
                    foreach ( $rtype->resources as $resource ) {
	                    //error_log( __METHOD__ . ' -- Going to import ' . $slug . ' "' . $resource->strTitle . '" which has ID ' . $resource->intID );

                        // determine post author and create if it doesn't exist
                        $author = array(
	                        'email' => $resource->strAuthorEmail,
	                        'first' => $resource->strAuthorFirstName,
	                        'last' => $resource->strAuthorLastName,
                        );
                        $author_id = $this->bt_vy_import_get_author( $author );

                        // build up WP post data based on the VY resource data -- keys and details come from WP developer documentation:
                        // https://developer.wordpress.org/reference/functions/wp_insert_post/#parameters
	                    $post = array(

                            'ID' => $resource->intID,
                            // (int) The post ID. If equal to something other than 0, the post with that ID will be updated. Default 0.

                            'post_author' => $author_id,
                            // (int) The ID of the user who added the post. Default is the current user ID.

		                    /**
		                     * For rtypes with date range, like events, we need to use the Last Updated date for the
		                     * publish date, as the start/single and end dates are separate meta fields.
		                     */
                            'post_date' => ( $rtype->intUseDateRange ? $resource->strLastUpdatedDate : $resource->strDate ),
                            // (string) The date of the post. Default is the current time.

                            /* NA 'post_date_gmt' => ,*/
                            // (string) The date of the post in the GMT timezone. Default is the value of $post_date.

                            'post_content' => $resource->strDescription,
                            // (mixed) The post content. Default empty.

                            /* NA 'post_content_filtered' => ,*/
                            // (string) The filtered post content. Default empty.

                            'post_title' => $resource->strTitle,
                            // (string) The post title. Default empty.

                            'post_excerpt' => $resource->strOverview,
                            // (string) The post excerpt. Default empty.

                            'post_status' => 'publish',
                            // (string) The post status. Default 'draft'.

                            'post_type' => $post_type_slug,
                            // (string) The post type. Default 'post'.

                            'comment_status' => 'closed',
                            // (string) Whether the post can accept comments. Accepts 'open' or 'closed'. Default is the value of 'default_comment_status' option.

                            'ping_status' => 'closed',
                            // (string) Whether the post can accept pings. Accepts 'open' or 'closed'. Default is the value of 'default_ping_status' option.

                            'post_password' => '',
                            // (string) The password to access the post. Default empty.

                            'post_name' => $resource->strSlug,
                            // (string) The post name. Default is the sanitized post title when creating a new post.

                            'to_ping' => '',
                            // (string) Space or carriage return-separated list of URLs to ping. Default empty.

                            'pinged' => '',
                            // (string) Space or carriage return-separated list of URLs that have been pinged. Default empty.

                            'post_modified' => $resource->strLastUpdatedDate,
                            /* @todo Why doesn't this work? Do we need to manually set it later? */
                            // (string) The date when the post was last modified. Default is the current time.

                            /* NA 'post_modified_gmt' => ,*/
                            // (string) The date when the post was last modified in the GMT timezone. Default is the current time.

                            'post_parent' => $resource->intParentRID,
                            // (int) Set this for the post it belongs to, if any. Default 0.

                            'menu_order' => ( is_numeric( $resource->intSortOrder ? $resource->intSortOrder : 0 ) ),
                            // (int) The order the post should be displayed in. Default 0.

                            'post_mime_type' => '',
                            // (string) The mime type of the post. Default empty.

                            'guid' => '',
                            // (string) Global Unique ID for referencing the post. Default empty.

                            'post_category' => array(),
                            // (array) Array of category names, slugs, or IDs. Defaults to value of the 'default_category' option.

                            'tax_input' => array(),
                            // (array) Array of taxonomy terms keyed by their taxonomy name. Default empty.

                            'meta_input' => array(),
                            // (array) Array of post meta values keyed by their post meta key. Default empty.

                        );

	                    // if post doesn't already exist, we need to use a different key in the post array so that the WP post ID matches the VY resource ID
                        if ( get_post_status( $resource->intID ) == false ) {
                            unset( $post['ID'] );
	                        $post['import_id'] = $resource->intID;
                        }
	                    // active, temp/pending, restricted
                        if ( !$resource->boolActive ) {
                            $post['post_status'] = 'draft';
                        }
	                    if ( $resource->intTempStatusID ) {
                            error_log( __METHOD__ . ' -- Resource has TEMP STATUS ID ' . $resource->intTempStatusID . '! Resource is "' . $resource->strTitle . '" which has ID ' . $resource->intID );
		                    $post['post_status'] = 'pending';
	                    }
	                    if ( $resource->boolRestricted ) {
		                    $post['post_status'] = 'private';
	                    }
	                    // title as person name
                        if ( $rtype->intTitleAsPersonName ) {
                            $post['post_title'] = $resource->strTitleAsPersonName;
                        }
                        // overview vs desc
                        if ( $rtype->intUseOverview  && !$rtype->intUseDesc ) {
                            // use the overview as the content rather than the excerpt since that's all there is
                            $post['post_content'] = $resource->strOverview;
                            $post['post_excerpt'] = '';
                        }
	                    // set categories
	                    if ( $rtype->intUseCategories && $tax_slug && $resource->intCategoryID ) {
	                        // get the term ID from the term we stored in the categories array above (with hierarchical taxonomy, you must use term ID rather than slug)
	                        $term_id = $rtype_categories[$resource->intCategoryID]['term']['term_taxonomy_id'];
	                        $post['tax_input'][$tax_slug] = array( $term_id );
                        }
                        // set meta
	                    foreach ( $this->cpt['meta'] as $meta_field => $meta_data) {
                            // skip any fields that need some additional processing
                            if ( in_array( $meta_data['name'], array( 'phone-numbers', 'images', 'file' ) ) ) {
                                continue;
                            }
                            // check if we're mapping to a field that goes multiple levels deep, like arbitrary attributes
                            if ( strpos( $meta_data['vy_map_to'], '->' ) ) {
                                // need to go multiple levels deep
	                            $value = NULL;
	                            $path = explode( '->', $meta_data['vy_map_to'] );
	                            // get first-level node
	                            $node = $resource->{ array_shift( $path ) };
	                            // loop through subsequent nodes until they end
	                            while ( ( $prop = array_shift( $path ) ) !== NULL ) {
		                            if ( ! $node->$prop && $node->$prop !== '' ) {
			                            break;
		                            }
		                            $value = $node->$prop;
		                            $node = $node->$prop;
	                            }
	                            // store final value
	                            $post['meta_input'][$meta_data['name']] = $value;

	                            // special cases for manipulation of other post data based on this meta
	                            $video_url = '';
	                            switch ( $meta_data['name'] ) {
		                            // construct video URLs and append to post content
	                                case 'youtube-id':
	                                    if ( $value ) $video_url = 'https://www.youtube.com/watch?v=' . $value;
	                                    break;
	                                case 'vimeo-id':
		                                if ( $value ) $video_url = 'https://vimeo.com/' . $value;
			                            break;
	                            }
	                            if ( $video_url ) {
	                                // append video URL to post content, including a couple of line breaks if there is content in there already
		                            $post['post_content'] .= ( $post['post_content'] ? PHP_EOL . PHP_EOL : '' ) . $video_url;
                                }

                            } else {
                                // direct mapping
	                            $post['meta_input'][$meta_data['name']] = $resource->{ $meta_data['vy_map_to'] };
                            }
	                    }

                        // insert the post
	                    $result = wp_insert_post( $post, true );
	                    if ( is_a( $result, 'WP_Error' ) ) {
		                    error_log( __METHOD__ . ' -- Inserting Post "' . $resource->strTitle . '" resulted in WP Error: ' . $result->get_error_message() );
	                    } else {
	                        // we successfully inserted so get the post's ID for subsequent use
		                    $inserted_post_id = $result;

		                    // add phone numbers
                            if ( $rtype->intUsePhoneNumbers && count( $resource->arrPhoneNumbers ) ) {
                                $field_key = 'field_' . $this->cpt_meta_slug_base . 'phone-numbers';
                                $phones = array();
                                foreach ( $resource->arrPhoneNumbers as $phone ) {
                                    $phones[] = array(
                                        $phone_sub_field_slug_base . 'label' => $phone->strTitle,
                                        $phone_sub_field_slug_base . 'area' => $phone->strArea,
                                        $phone_sub_field_slug_base . 'exchange' => $phone->strExchange,
                                        $phone_sub_field_slug_base . 'suffix' => $phone->strSuffix,
                                        $phone_sub_field_slug_base . 'ext' => $phone->strExtension,
                                    );
                                }
                                update_field( $field_key, $phones, $inserted_post_id );
                            }

                            // add main image
                            if ( $rtype->intUseImage ) {
                                // check for main image in this resource
                                if ( $resource->strImageFilename ) {
	                                $attachment_args = array(
		                                'post_id' => $inserted_post_id,
		                                'vy_file_field' => 'strImageFilename',
		                                'vy_file_type' => 'Image',
		                                'rtype_name' => $rtype->strName,
	                                );
                                    $attachment_id = $this->bt_vy_import_add_post_attachment( $resource, $attachment_args );
                                }
                            }

		                    // add download/document
                            if ( $this->bt_vy_import_rtype_has_file( $slug, $rtype->intUseDownload ) ) {
			                    // check for download in this resource
			                    if ( $resource->strDownloadFilename ) {
				                    $attachment_args = array(
					                    'post_id' => $inserted_post_id,
					                    'vy_file_field' => 'strDownloadFilename',
					                    'vy_file_type' => 'Download',
					                    'rtype_name' => $rtype->strName,
				                    );
				                    $attachment_id = $this->bt_vy_import_add_post_attachment( $resource, $attachment_args );
				                    // update the custom meta field
				                    $field_key = 'field_' . $this->cpt_meta_slug_base . 'file';
				                    update_field( $field_key, $attachment_id, $inserted_post_id );
			                    }
		                    }

		                    // add gallery images
		                    if ( $rtype->intUseMultiplePhotos && count( $resource->arrGalleryImages ) ) {
			                    $field_key = 'field_' . $this->cpt_meta_slug_base . 'images';
			                    $images = array();
			                    foreach ( $resource->arrGalleryImages as $image ) {
			                        // create attachment for current image
				                    $attachment_args = array(
					                    'post_id' => $inserted_post_id,
					                    'vy_file_name' => $image->strImageFilename,
					                    'vy_file_type' => 'Image',
					                    'rtype_name' => $rtype->strName,
                                        'post_date' => $image->strLastUpdated,
				                    );
				                    $attachment_id = $this->bt_vy_import_add_post_attachment( $resource, $attachment_args );
			                        // add image attachment ID to our meta field array
				                    $images[] = $attachment_id;
			                    }
			                    update_field( $field_key, $images, $inserted_post_id );
		                    }

	                    }

                    }

					$current_rtype_imported = true;
				}

				/** ****************************************************************************************************
				 * OUTPUT UI for current rtype
				 **************************************************************************************************** */
				?>
					<div id="rtype-<?php echo $rtype->intID; ?>" class="rtype<?php echo $current_rtype_imported ? ' vy_success' : ''; ?>">
						<h4>
							<input type="checkbox" name="bt_vy_rtypes[]" value="<?php echo $rtype->intID; ?>" id="rtype-chk-<?php echo $rtype->intID; ?>" <?php if ( in_array( $rtype->intID, $rtypes_to_import ) ) echo ' checked'; ?>>
							<label for="rtype-chk-<?php echo $rtype->intID; ?>"><?php echo $this->cpt['single']; ?> (<?php echo $rtype->resource_count; ?>)</label>
						</h4>
						<div class="rtype_details">
							<section>
                                <?php
                                // loop fields to output
                                foreach ( $editable_fields as $field_slug => $field_data ) {
                                    $this->bt_vy_import_rtype_field_markup( $rtype->intID, $field_slug, $field_data['label'], $field_data['value'] );
                                }
                                ?>
                            </section>
                            <p class="debug-trigger">Toggle Debug</p>
                            <section class="debug">
                                <textarea class="vy_cpt_data-<?php echo $rtype->intID; ?>"><?php echo json_encode( $this->cpt ); ?></textarea>
                                <p>Meta:</p>
                                <pre><?php print_r( $this->cpt['meta'] ); ?></pre>
                                <p>Options:</p>
                                <pre><?php print_r( $this->cpt['options'] ); ?></pre>
                            </section>
						</div>
                    </div>
				<?php

			}

			// if we did the full import and made it here, take final post-import actions
			if ( $run_import ) {
			    // reset the "run import" checkbox option and set a timestamp
			    $import_time = current_time( 'Y-m-d H:i:s' );
				update_option( $this->base . 'vy_import_run', '' );
				update_option( $this->base . 'vy_last_import', $import_time );
				// output markup to un-check the checkbox and set the last import time since they will have been previous values when the page loaded
                ?>
                <script type="text/javascript">
                    jQuery(document).ready(function($) {
                       $('#vy_import_run').prop('checked', false);
                       $('#last_import_time').html('<?php echo $import_time; ?>');
                    });
                </script>
                <?php
            }

			// make sure our cpt paths are set up in rewrites
			flush_rewrite_rules();

		} else {
			echo 'Bad response from URI!<br>URI: ' . $json_uri . '<br>RESPONSE: <pre>' . print_r( $api['response'], 1 ) . '</pre>';
			wp_die();
		}

		/**
		 * @todo NEXT STEPS:
         * - Documentation of adding new fields, etc. [SEE EMAIL DRAFT]
         * - Links: remove default unusable meta widget for "link categories" which appears via some default WP functionality with post_type=link
         * - Custom capabilities to restrict in cases of locked categories.
         * - Hide Custom Fields menu @see https://www.advancedcustomfields.com/resources/how-to-hide-acf-menu-from-clients/
         *
         * AFTER BFO:
         * - Extra Images for rtypes (BFO only used for videos/galleries and that was for old sidebar features)
         * - Gallery Images: bring in title and description (BFO didn't use these fields)
         * - Products and product categories
		 */

		wp_die();
	} // bt_vy_import()


	/**
	 * Generate markup for a resource type field
	 *
	 * @since 2.0.1
     * @private
     * @param $rtype_id     int     The ID of the current VenYou Resource Type
     * @param $field_slug   string  The slug that identifies the current field
     * @param $field_label  string  The proper English name/label for the field
     * @param $field_value  mixed   The current value for the field
     * @return void
	 */
	private function bt_vy_import_rtype_field_markup( $rtype_id, $field_slug, $field_label, $field_value ) {
		$selector = 'rtype-' . $field_slug . '-' . $rtype_id;
	    ?>
        <p>
            <label for="<?php echo $selector; ?>">
                <?php echo $field_label; ?>
            </label>
            <input type="text" name="<?php echo $selector; ?>" id="<?php echo $selector; ?>" value="<?php echo $field_value; ?>" data-rtype-id="<?php echo $rtype_id; ?>" data-field-slug="<?php echo $field_slug; ?>">
        </p>
        <?php
	} // bt_vy_import_rtype_field_markup()

	/**
	 * Generate an array of ACF Pro field data which corresponds with some rtype/cpt meta field
	 *
	 * @since 2.0.1
	 * @private
     * @param $field_title string The proper English name of the field
     * @param $field_type string The type of ACF field; @see https://www.advancedcustomfields.com/resources/#field-types
     * @param $args array Any field elements to be set beyond the defaults
	 * @return void
	 */
	private function bt_vy_import_rtype_add_field_meta( $field_title, $field_type = 'text', $args = array() ) {

	    // determine meta name and full slug based on the given title
		$meta_name = sanitize_title( $field_title ); // used as the ACF field name (what you'd call to retrieve the data on front end)
		$meta_slug = $this->cpt_meta_slug_base . $meta_name; // used as the ACF field key

		// set up any defaults dependent on type
        switch ( $field_type ) {
            case 'email':
                $placeholder = 'example@domain.com';
                break;
	        case 'url':
		        $placeholder = 'https://example.com/path/to/content';
		        break;
            default:
	            $placeholder = '';
        }

		// merge arguments with defaults, and add it to all CPT meta for return
		$this->cpt['meta'][$meta_slug] = array_merge(
			array(
				'label' => __( $field_title, 'bt-tools' ),
				'name' => $meta_name,
				'type' => $field_type,
				'placeholder' => $placeholder,
				'maxlength' => '',
			),
			$args
		);

        //error_log( __METHOD__ . ' -- Current field data for slug ' . $meta_slug . ' is: ' . print_r( $this->cpt['meta'][$meta_slug], 1 ) );

	} // bt_vy_import_rtype_add_field_meta()

	/**
	 * Remove a "supports" option from the current CPT
	 *
	 * @since 2.0.1
	 * @private
	 * @param $option string The name of an option to remove, such as "editor" or "thumbnail"
	 * @return void
	 */
	private function bt_vy_import_rtype_remove_supports( $option ) {

		$this->cpt['options']['supports'] = array_diff(
			$this->cpt['options']['supports'],
			array( $option )
		);

	} // bt_vy_import_rtype_remove_supports()


	/**
     * Determine whether a resource type should have the custom meta field for file.
     *
	 * @param $slug string The slug of the resource type we are testing.
	 * @param $use_download int The flag from the rtype config which says whether the rtype uses a download or not.
     *
	 * @return bool Whether the rtype should use the file custom meta field.
	 */
    public function bt_vy_import_rtype_has_file( $slug, $use_download ) {
	    // make sure rtype uses a download
	    if ( ! $use_download ) {
	        return false;
        }
	    // return false for any rtypes that use a file but we don't want it to carry forward here
	    switch ( $slug ) {
            case 'video': // old FLVs
            case 'image-gallery': // zip files for uploading multiple at once
                return false;
        }
        // if we're here, it must use download and we want to use it as a file in WP
        return true;
    }


    /**
     * Add an attachment for an imported post, such as the resource image or download.
     *
	 * @param $resource stdClass The resource object which is currently being processed
	 * @param $args array Bits of info we need for creating an attachment: post ID, file field, file type, rtype name.
     *
     * @return mixed Numeric ID of successfully-inserted attachment, else false.
	 */
    private function bt_vy_import_add_post_attachment( $resource, $args ) {

	    $args = array_merge(
            array(
                'post_id' => 0,
	            'vy_file_type' => 'Image',
                'vy_file_field' => 'strImageFilename', // if set, this will get the filename from a $resource data member
                'vy_file_name' => '', // if set, this specific filename will be used instead of the $resource data member. used for cases like gallery images where data is deep.
                'rtype_name' => '',
	            'post_date' => '', // if set, this specific timestamp will be used instead of the $resource image updated timestamp. used for cases like gallery images where data is deep.
            ),
            $args
        );

	    // build path to file and make sure it exists
        if ( $args['vy_file_name'] ) {
	        $file_name = $args['vy_file_name'];
        } else {
	        $file_name = $resource->$args['vy_file_field'];
        }
	    $file_source = $resource->strFileRoot . '_uls/resources/' . $file_name;
	    if ( file_exists( $file_source ) ) {

		    // copy file into wp uploads
		    $file_dest = $this->import_file_path . $file_name;
		    if ( ! copy( $file_source, $file_dest ) ) {
			    error_log( __METHOD__ . ' -- Inserted Post "' . $resource->strTitle . '" but could not copy ' . $args['vy_file_type'] . ': ' . $file_source );
		    }
		    $file_type = wp_check_filetype( basename( $file_dest ) );

		    // build up attachment data
		    $file_args = array(
			    'guid' => $this->import_file_url . $file_name,
			    'post_mime_type' => $file_type['type'],
			    'post_date' => ( $args['post_date'] ? $args['post_date'] : $resource->strImageFileLastUpdatedDate ),
			    'post_title' => $args['vy_file_type'] . ' for ' . $args['rtype_name'] . ': ' . $resource->strTitle,
			    'post_excerpt' => '', // goes into "caption" field
			    'post_content' => '', // goes into "description" field
			    'post_status' => 'inherit',
			    'comment_status' => 'closed',
			    'ping_status' => 'closed',
		    );

		    // determine if we should specify the ID or not. we do if it hasn't been done yet.
		    // if it has been done, then $this->first_available_id is null and we can let WP generate IDs.
		    if ( $this->first_available_id ) {
			    $file_args['import_id'] = $this->first_available_id;
			    // reset it as we've now specified the first ID and WP will increment from there.
			    $this->first_available_id = null;
		    }

		    // insert the attachment
		    $result = wp_insert_attachment( $file_args, $file_dest, $args['post_id'] );
		    if ( is_a( $result, 'WP_Error' ) ) {
			    error_log( __METHOD__ . ' -- Inserting ' . $args['vy_file_type'] . ' for Post "' . $resource->strTitle . '" resulted in WP Error: ' . $result->get_error_message() );
		    } else {
			    // we successfully inserted so get the attachment's ID for subsequent use
			    $inserted_attachment_id = $result;

			    // generate the metadata for the attachment, and update the database record
			    require_once( ABSPATH . 'wp-admin/includes/image.php' );
			    $attach_data = wp_generate_attachment_metadata( $inserted_attachment_id, $file_dest );
			    wp_update_attachment_metadata( $inserted_attachment_id, $attach_data );

			    // image-specific items
                if ( $args['vy_file_type'] == 'Image' ) {

	                // set this as the post thumbnail
	                set_post_thumbnail( $args['post_id'], $inserted_attachment_id );

	                // set the image meta
	                update_post_meta( $inserted_attachment_id, '_wp_attachment_image_alt', $resource->strImageAltText );

                }

                // return the successfully-inserted attachment id
                return $inserted_attachment_id;

		    }

	    } else {
		    error_log( __METHOD__ . ' -- Inserted Post "' . $resource->strTitle . '" but its ' . $args['vy_file_type'] . ' does not exist: ' . $file_source );
	    }

	    // something failed
        return false;

    }


	/**
	 * Determine the WordPress author given some author data. Will insert a new author if it can't find an existing one.
     * Will use current author if given data is "unknown."
	 *
	 * @param $author array The email, first name, and last name of an author.
	 *
	 * @return int The ID of the corresponding author.
	 */
	public function bt_vy_import_get_author( $author ) {
	    if ( ! $author['email'] || $author['email'] == 'unknown' ) {
	        // we don't have an author; the current user will have to do
            return get_current_user_id();
        } else {
	        // look for an existing author by email address
		    $user = get_user_by( 'email', $author['email'] );
		    if ( ! empty( $user ) ) {
		        // we found an existing user, so return its ID
                return $user->ID;
            } else {
			    // we need to create a user
			    $email_parts = explode('@', $author['email'] );
			    $login = $email_parts[0];
			    $userdata = array(
				    'user_pass' => wp_generate_password( 20 ),
				    'user_login' => $login,
				    'user_nicename' => $login,
				    'user_email' => $author['email'],
				    'first_name' => $author['first'],
				    'last_name' => $author['last'],
				    'role' => 'author',
			    );
			    $result = wp_insert_user($userdata);
			    if (is_a($result, 'WP_Error')) {
				    // log error and default to using the current user
				    error_log( __METHOD__ . ' -- Could not insert author ' . $author['email'] . '! WP_Error message: ' . $result->get_error_message() );
				    return get_current_user_id();
			    } else {
			        // success - the result is the new user ID
                    return $result;
                }
		    }
        }
	}


}
