<?php

if ( ! defined( 'ABSPATH' ) ) exit;

class BT_Tools {

	/**
	 * The single instance of BT_Tools.
	 * @var 	object
	 * @access  private
	 * @since 	1.0.0
	 */
	private static $_instance = null;

	/**
	 * Settings class object
	 * @var     object
	 * @access  public
	 * @since   1.0.0
	 */
	public $settings = null;

	/**
	 * The version number.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $_version;

	/**
	 * The token.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $_token;

	/**
	 * The main plugin file.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $file;

	/**
	 * The main plugin directory.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $dir;

	/**
	 * The plugin assets directory.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $assets_dir;

	/**
	 * The plugin assets URL.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $assets_url;

	/**
	 * Suffix for Javascripts.
	 * @var     string
	 * @access  public
	 * @since   1.0.0
	 */
	public $script_suffix;

	/**
	 * Meta fields for custom post types from VenYou.
	 * @var     array
	 * @access  public
	 * @since   2.0.1
	 */
	public $vy_post_type_meta;

	/**
	 * Constructor function.
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function __construct ( $file = '', $version = '1.0.0' ) {
		$this->_version = $version;
		$this->_token = 'bt_tools';

		// Load plugin environment variables
		$this->file = $file;
		$this->dir = dirname( $this->file );
		$this->assets_dir = trailingslashit( $this->dir ) . 'assets';
		$this->assets_url = esc_url( trailingslashit( plugins_url( '/assets/', $this->file ) ) );

		$this->script_suffix = defined( 'SCRIPT_DEBUG' ) && SCRIPT_DEBUG ? '' : '.min';

		$this->vy_post_type_meta = array();

		register_activation_hook( $this->file, array( $this, 'install' ) );

		// initialize custom post types from VenYou
		add_action( 'init', array( $this, 'vy_init_post_types' ), 0);
		add_action( 'acf/init', array( $this, 'vy_init_post_type_meta' ) );

		// save and load ACF data (user-defined; not meta from VenYou rtypes)
		//add_filter('acf/settings/save_json', array( $this, 'bt_acf_json_save_point' ) );
		//add_filter('acf/settings/load_json', array( $this, 'bt_acf_json_load_point' ) );

		// Load frontend JS & CSS
		/**
		 * We're not using front-end css and js
		add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_styles' ), 10 );
		add_action( 'wp_enqueue_scripts', array( $this, 'enqueue_scripts' ), 10 );
		 */

		// Load admin JS & CSS
		add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_scripts' ), 10, 1 );
		add_action( 'admin_enqueue_scripts', array( $this, 'admin_enqueue_styles' ), 10, 1 );

		// add action for output of webfonts config/loading markup
		/** @todo Resolve whether we want to include this or not.
		 *          For bfo/w4 we are using the webfontloader via npm and including font config in js that gets compiled/minified.
		 *          The mechanism here could be used with the options we're saving, but it would need to load in the webfontloader js (see tru/ran) and we wouldn't do this with a Sage base.
		 * if ( ! is_admin() ) {
			add_action( 'wp_head' , array( $this, 'bt_fonts_loader' ), 9999);
		}*/

		// create shortcodes
		add_shortcode( 'bt_rss_feed', array( $this, 'bt_rss_feed' ) );

		// Load API for generic admin functions
		if ( is_admin() ) {
			$this->admin = new BT_Tools_Admin_API();
			// add separator before custom rtypes
			add_action( 'admin_init', array( $this, 'vy_admin_menu_update' ) );
			// add custom taxonomy filters to custom post types
			add_action( 'restrict_manage_posts', array( $this, 'vy_filter_post_type_by_taxonomy' ) );
			// filter posts by custom taxonomy
			add_action( 'parse_query', array( $this, 'vy_convert_id_to_term_in_query' ) );
		}

		// Handle localisation
		$this->load_plugin_textdomain();
		add_action( 'init', array( $this, 'load_localisation' ), 0 );
	} // End __construct ()

	/**
	 * Wrapper function to register a new post type
	 * @param  string $post_type   Post type name
	 * @param  string $plural      Post type item plural name
	 * @param  string $single      Post type item single name
	 * @param  string $description Description of post type
	 * @return object              Post type class object
	 */
	public function register_post_type ( $post_type = '', $plural = '', $single = '', $description = '', $options = array() ) {

		if ( ! $post_type || ! $plural || ! $single ) return;

		$post_type = new BT_Tools_Post_Type( $post_type, $plural, $single, $description, $options );

		return $post_type;
	}

	/**
	 * Wrapper function to register a new taxonomy
	 * @param  string $taxonomy   Taxonomy name
	 * @param  string $plural     Taxonomy single name
	 * @param  string $single     Taxonomy plural name
	 * @param  array  $post_types Post types to which this taxonomy applies
	 * @return object             Taxonomy class object
	 */
	public function register_taxonomy ( $taxonomy = '', $plural = '', $single = '', $post_types = array(), $taxonomy_args = array() ) {

		if ( ! $taxonomy || ! $plural || ! $single ) return;

		$taxonomy = new BT_Tools_Taxonomy( $taxonomy, $plural, $single, $post_types, $taxonomy_args );

		return $taxonomy;
	}

	/**
	 * Load frontend CSS.
	 * @access  public
	 * @since   1.0.0
	 * @return void
	 */
	public function enqueue_styles () {
		wp_register_style( $this->_token . '-frontend', esc_url( $this->assets_url ) . 'css/frontend.css', array(), $this->_version );
		wp_enqueue_style( $this->_token . '-frontend' );
	} // End enqueue_styles ()

	/**
	 * Load frontend Javascript.
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function enqueue_scripts () {
		wp_register_script( $this->_token . '-frontend', esc_url( $this->assets_url ) . 'js/frontend' . $this->script_suffix . '.js', array( 'jquery' ), $this->_version );
		wp_enqueue_script( $this->_token . '-frontend' );
	} // End enqueue_scripts ()

	/**
	 * Load admin CSS.
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function admin_enqueue_styles ( $hook = '' ) {
		wp_register_style( $this->_token . '-admin', esc_url( $this->assets_url ) . 'css/admin.css', array(), $this->_version );
		wp_enqueue_style( $this->_token . '-admin' );
	} // End admin_enqueue_styles ()

	/**
	 * Load admin Javascript.
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function admin_enqueue_scripts ( $hook = '' ) {
		wp_register_script( $this->_token . '-admin', esc_url( $this->assets_url ) . 'js/admin' . $this->script_suffix . '.js', array( 'jquery' ), $this->_version );
		wp_enqueue_script( $this->_token . '-admin' );
	} // End admin_enqueue_scripts ()

	/**
	 * Load plugin localisation
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function load_localisation () {
		load_plugin_textdomain( 'bt-tools', false, dirname( plugin_basename( $this->file ) ) . '/lang/' );
	} // End load_localisation ()

	/**
	 * Load plugin textdomain
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function load_plugin_textdomain () {
	    $domain = 'bt-tools';

	    $locale = apply_filters( 'plugin_locale', get_locale(), $domain );

	    load_textdomain( $domain, WP_LANG_DIR . '/' . $domain . '/' . $domain . '-' . $locale . '.mo' );
	    load_plugin_textdomain( $domain, false, dirname( plugin_basename( $this->file ) ) . '/lang/' );
	} // End load_plugin_textdomain ()

	/**
	 * Main BT_Tools Instance
	 *
	 * Ensures only one instance of BT_Tools is loaded or can be loaded.
	 *
	 * @since 1.0.0
	 * @static
	 * @see BT_Tools()
	 * @return Main BT_Tools instance
	 */
	public static function instance ( $file = '', $version = '1.0.0' ) {
		if ( is_null( self::$_instance ) ) {
			self::$_instance = new self( $file, $version );
		}
		return self::$_instance;
	} // End instance ()

	/**
	 * Cloning is forbidden.
	 *
	 * @since 1.0.0
	 */
	public function __clone () {
		_doing_it_wrong( __FUNCTION__, __( 'Cheatin&#8217; huh?' ), $this->_version );
	} // End __clone ()

	/**
	 * Unserializing instances of this class is forbidden.
	 *
	 * @since 1.0.0
	 */
	public function __wakeup () {
		_doing_it_wrong( __FUNCTION__, __( 'Cheatin&#8217; huh?' ), $this->_version );
	} // End __wakeup ()

	/**
	 * Installation. Runs on activation.
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	public function install () {
		$this->_log_version_number();
	} // End install ()

	/**
	 * Log the plugin version number.
	 * @access  public
	 * @since   1.0.0
	 * @return  void
	 */
	private function _log_version_number () {
		update_option( $this->_token . '_version', $this->_version );
	} // End _log_version_number ()


	/**
	 * Gets the VenYou resource type config from options.
	 * @access  public
	 * @since   2.0.1
	 * @return  object JSON-decoded stdClass of the VenYou resource types that have been selected for import
	 */
	public function vy_get_rtypes ( $as_array = false ) {
		// get rtype config from database
		return json_decode( urldecode( get_option( 'bt_vy_rtypes_to_import' ) ), $as_array );
	} // End vy_get_rtypes ()

	/**
	 * Registers custom post types based on the VenYou resource type config in options.
	 * @access  public
	 * @since   2.0.1
	 * @return  void
	 */
	public function vy_init_post_types () {

		// get rtype config from database
		$rtypes = $this->vy_get_rtypes( true ); // true param returns data as array instead of stdClass

		// loop rtypes
		foreach ( $rtypes as $rtype ) {

			//error_log( __METHOD__ . ' -- ' . $rtype['slug'] . ' -- ' . print_r( $rtype, 1) );
			$this->register_post_type(
				$rtype['slug'],
				__( $rtype['plural'], 'bt-tools' ),
				__( $rtype['single'], 'bt-tools' ),
				'',
				$rtype['options']
			);

			// create any taxonomy if exists
			if ( count( $rtype['options']['taxonomies_extended'] ) ) {
				foreach ( $rtype['options']['taxonomies_extended'] as $taxonomy => $taxonomy_data ) {
					//error_log( __METHOD__ . ' -- Taxonomy ' . $taxonomy . ' -- ' . print_r( $taxonomy_data, 1) );
					$this->register_taxonomy(
						$taxonomy,
						__( $taxonomy_data['plural'], 'bt-tools' ),
						__( $taxonomy_data['single'], 'bt-tools' ),
						array( $rtype['slug'] ),
						$taxonomy_data['args']
					);
				}
			}

			// create Advanced Custom Fields from meta if exists
			if ( count( $rtype['meta'] ) ) {

				// set up a field group for custom VenYou meta
				$field_group_slug = 'group_vy_core_' . $rtype['slug'];
				$this->vy_post_type_meta[$field_group_slug] = array(
					'key' => $field_group_slug,
					'title' => $rtype['single'] . ' Meta',
					'fields' => array(),
					'location' => array(
						array(
							array(
								'param' => 'post_type',
								'operator' => '==',
								'value' => $rtype['slug'],
							),
						),
					),
					'menu_order' => 0,
				);

				// set up a field group for related content
				$field_group_slug_related = 'group_vy_related_' . $rtype['slug'];
				$this->vy_post_type_meta[$field_group_slug_related] = array(
					'key' => $field_group_slug_related,
					'title' => 'Related Content',
					'fields' => array(),
					'location' => array(
						array(
							array(
								'param' => 'post_type',
								'operator' => '==',
								'value' => $rtype['slug'],
							),
						),
					),
					'menu_order' => 1,
				);

				// set up a field group for expiration date
				$field_group_slug_expiration = 'group_vy_expiration_' . $rtype['slug'];
				$this->vy_post_type_meta[$field_group_slug_expiration] = array(
					'key' => $field_group_slug_expiration,
					'title' => $rtype['single'] . ' Expiration',
					'fields' => array(),
					'location' => array(
						array(
							array(
								'param' => 'post_type',
								'operator' => '==',
								'value' => $rtype['slug'],
							),
						),
					),
					'menu_order' => 0,
					'position' => 'side'
				);

				// loop fields and add each to a group
				foreach ( $rtype['meta'] as $meta_field => $meta_data ) {
					//error_log( __METHOD__ . ' -- Meta ' . $meta_field . ' -- ' . print_r( $meta_data, 1) );

					// add a key to the meta data for ACF to use
					$meta_data['key'] = 'field_' . $meta_field;

					// add this meta field to the group fields
					if ( strpos( $meta_field, 'related' ) !== false ) {
						// special handling (separate group in main area) for related content
						$this->vy_post_type_meta[$field_group_slug_related]['fields'][] = $meta_data;

					} else if ( strpos( $meta_field, 'expiration-date' ) !== false ) {
						// special handling (group at side below publish box) for expiration date
						$this->vy_post_type_meta[$field_group_slug_expiration]['fields'][] = $meta_data;

					} else {
						// standard field for vy meta group
						$this->vy_post_type_meta[$field_group_slug]['fields'][] = $meta_data;

					}
				}
			}

		}

	} // End vy_init_post_types ()

	/**
	 * Registers custom post type meta fields within Advanced Custom Fields.
	 * @access  public
	 * @since   2.0.1
	 * @return  void
	 */
	public function vy_init_post_type_meta () {
		// make sure ACF is available
		if ( function_exists( 'acf_add_local_field_group' ) ) {
			// loop meta data previously set up in the class instance by method vy_init_post_types()
			foreach ( $this->vy_post_type_meta as $field_group ) {
				//error_log( __METHOD__ . ' -- Loading field group ' . $field_group['key'] . ':' . print_r( $field_group, 1 ) );
				if ( count( $field_group['fields'] ) ) {
					acf_add_local_field_group( $field_group );
				}
			}
		}
	} // End vy_init_post_type_meta ()

	/**
	 * Add custom taxonomy filters to custom post types
	 * @since 2.0.1
	 * @return void
	 */
	public function vy_filter_post_type_by_taxonomy() {
		global $typenow;
		$rtypes = $this->vy_get_rtypes();
		// loop until we find the matching post type slug (rtypes has VY rtype IDs as keys so we can't reference post type slug directly)
		foreach ( $rtypes as $rtype ) {
			if ( $rtype->slug == $typenow ) {
				// we found the rtype that matches the current post type; check for taxonomies
				if ( count( $rtype->options->taxonomies ) ) {
					// it has custom taxonomies so let's use them
					foreach ( $rtype->options->taxonomies as $taxonomy => $taxonomy_data ) {
						$selected = isset( $_GET[$taxonomy] ) ? $_GET[$taxonomy] : '';
						$info_taxonomy = get_taxonomy( $taxonomy );
						wp_dropdown_categories(array(
							'show_option_all' => __( "Show All {$info_taxonomy->label}" ),
							'taxonomy'        => $taxonomy,
							'name'            => $taxonomy,
							'orderby'         => 'name',
							'selected'        => $selected,
							'show_count'      => true,
							'hide_empty'      => true,
						));
					}
				}
				// exit our for loop since we found what we needed
				break;
			}
		}
	} // End vy_filter_post_type_by_taxonomy ()

	/**
	 * Filters posts by custom taxonomy in admin
	 * @since 2.0.1
	 * @param WP_Query $query The query which we're going to adjust
	 * @return void
	 */
	public function vy_convert_id_to_term_in_query( $query ) {
		global $pagenow;
		$q_vars = &$query->query_vars;
		if ( $pagenow == 'edit.php' && isset($q_vars['post_type']) ) {
			// we're on the edit page and there's a post type set in the query vars - proceed with processing
			$rtypes = $this->vy_get_rtypes();
			// loop until we find the matching post type slug (rtypes has VY rtype IDs as keys so we can't reference post type slug directly)
			foreach ( $rtypes as $rtype ) {
				if ( $rtype->slug == $q_vars['post_type'] ) {
					// we found the rtype that matches the current post type; check for taxonomies
					if ( count( $rtype->options->taxonomies ) ) {
						// it has custom taxonomies so let's use them
						foreach ( $rtype->options->taxonomies as $taxonomy => $taxonomy_data ) {
							if ( isset( $q_vars[$taxonomy] ) && is_numeric( $q_vars[$taxonomy] ) && $q_vars[$taxonomy] != 0 ) {
								$term = get_term_by( 'id', $q_vars[$taxonomy], $taxonomy );
								$q_vars[$taxonomy] = $term->slug;
							}
						}
					}
					// exit our for loop since we found what we needed
					break;
				}
			}
		}
	} // End vy_convert_id_to_term_in_query ()

	/**
	 * Adjust admin menu items as we need to
	 * @since 2.0.2
	 * @return void
	 */
	public function vy_admin_menu_update() {
		global $menu;
		// get rtype config from database
		$rtypes = $this->vy_get_rtypes( true ); // true param returns data as array instead of stdClass
		if ( count( $rtypes ) && is_array( $menu ) ) {

			// loop all menu items to adjust the ones we need to
			$positions_to_unset = [];
			foreach ( $menu as $position => $item ) {
				switch ( $item[2] ) {
					/**
					 * @since 2.0.6
					 * We didn't have posts before, but we now do.
					 */
					//case 'edit.php': // posts
					case 'edit-comments.php':
						// remove comments so long as we aren't using them
						// @todo Find a way to make this dynamic config for different sites
						$positions_to_unset[] = $position;
						break;
					case 'edit.php?post_type=article':
						// add a separator before post types from VenYou
						// @todo Find a way to make this NOT depend on Articles being the first rtype
						$where = (string)($position - 0.00001);
						$menu[$where] = array(
							'',							// The text of the menu item
							'read',						// Permission level required to view the item
							'separator-before-vycpt',   // The ID of the menu item
							'',							// Empty by default.
							'wp-menu-separator'			// Custom class names for the menu item
						);
						break;
					case 'wpcf7':
						// only process the original item - this hook runs more than once sometimes
						if ( is_int( $position ) ) {
							// move contact form 7 to just before Appearance menu, and put a separator after it
							$positions_to_unset[] = $position;
							$menu['59.501'] = $item;
							$menu['59.503'] = array(
								'',							// The text of the menu item
								'read',						// Permission level required to view the item
								'separator-after-wpcf7',   // The ID of the menu item
								'',							// Empty by default.
								'wp-menu-separator'			// Custom class names for the menu item
							);
						}
						break;
					case 'flamingo':
						// only process the original item - this hook runs more than once sometimes
						if ( is_int( $position ) ) {
							// move flamingo to just before Appearance menu, after contact form 7 but before its separator
							$positions_to_unset[] = $position;
							$menu['59.502'] = $item;
						}
						break;
					default:
						// ok as is
				}
			}

			// remove the items we want to get rid of
			foreach ( $positions_to_unset as $position_to_unset ) {
				unset( $menu[$position_to_unset] );
			}

			// re-sort everything
			ksort( $menu );
		}
	}

	/**
	 * Save ACF fields (user-defined; not those imported from VenYou) in JSON format here within plugin directory
	 * @since 2.0.1
	 * @param $path string The original path passed through the filter.
	 * @return string Path to location where JSON files will be stored
	 */
	public function bt_acf_json_save_point( $path ) {
		return trailingslashit( $this->dir ) . 'acf-json';
	}

	/**
	 * Load ACF fields (user-defined; not those imported from VenYou) in JSON format here within plugin directory
	 * @since 2.0.1
	 * @params $path array The original paths passed through the filter.
	 * @return array Paths to locations where JSON files will be stored
	 */
	public function bt_acf_json_load_point( $paths ) {
		// append our path (could remove default by unsetting $paths[0]
		$paths[] = trailingslashit( $this->dir ) . 'acf-json';
		return $paths;
	}

	/**
	 * Retrieves social link options; constructs and returns an array of them for processing in a theme
	 * @access  public
	 * @since   2.0.1
	 * @return  array Social links array, each containing type, text, and url
	 */
	public function bt_social_links () {
		// init and get options from database
		$links = [];
		for ( $i = 1; $i <= BT_SOCIAL_LINK_SLOTS; $i++ ) {
			$type = get_option( 'bt_social_link_'.$i.'_type' );
			$text = get_option( 'bt_social_link_'.$i.'_text' );
			$url = get_option( 'bt_social_link_'.$i.'_url' );
			if ( $type && $text && $url ) {
				$link = new stdClass;
				$link->type = $type;
				$link->text = $text;
				$link->url = $url;
				$links[] = $link;
			}
		}
		return $links;
	} // End bt_social_links ()



	/**
	 * Output font loader <script> and <noscript> tags. Does NOT load in google webfonts loader JS; it is assumed this is in place independently!
	 **/
	function bt_fonts_loader() {
		if ( get_option( 'bt_fonts_on' ) ) {
			// get fonts.com and google fonts data
			$fonts_com_project = get_option( 'bt_fonts_fontscom_project_id' );
			$google_fonts = get_option( 'bt_fonts_google' );
			// build options that are in use
			$wfl_config = '{';
			$noscript = '<noscript>';
			if ($fonts_com_project) {
				$wfl_config .= 'monotype:{projectId:"' . $fonts_com_project . '"},';
				$noscript .= '<link type="text/css" rel="stylesheet" href="https://fast.fonts.com/cssapi/' . $fonts_com_project . '.css">';
			}
			if ($google_fonts) {
				$wfl_config .= 'google:{families:[' . $this->bt_fonts_prepare_google_families( $google_fonts, true ) . ']},';
				$noscript .= '<link href="https://fonts.googleapis.com/css?family=' . $this->bt_fonts_prepare_google_families( $google_fonts, false ) . '" rel="stylesheet" type="text/css">';
			}
			$wfl_config = rtrim($wfl_config, ',') . '}';
			$noscript .= '</noscript>';
			// output
			echo '<script type="text/javascript">WebFont.load('.$wfl_config.');</script>' . PHP_EOL . $noscript;
		}
	}

	/**
	 * Prep google font families for use in web fonts loader config or stylesheet url
	 *
	 * @param $families string Google font families like "Droid Sans:bold", each separated by EOLs
	 * @param $for_wfl_config bool If true, construct markup for the WFL json config, else for the css family URL used in noscript
	 *
	 * @return string Prepared markup for use by callee
	 */
	function bt_fonts_prepare_google_families( $families, $for_wfl_config = true ) {
		$markup = '';
		foreach (explode(PHP_EOL, $families) as $family) {
			$family = preg_replace("/[\n\r]/","",$family);
			if ($for_wfl_config) {
				$markup .= "'$family',";
			} else {
				$markup .= str_replace(' ', '+', "$family|");
			}
		}
		if ($for_wfl_config) {
			$markup = rtrim($markup, ',');
		} else {
			$markup = rtrim($markup, '|');
		}
		return $markup;
	}

	/**
	 * Fetch info about a video from YouTube or Vimeo
	 */
	public static function bt_video_data( $id, $service = 'youtube' ) {
		if ( ! $id ) {
			return false;
		}
		switch ( $service ) {
			case 'youtube':
				$request = 'https://www.googleapis.com/youtube/v3/videos' .
				           '?id=' . $id .
				           '&key=' . YOUTUBE_API_KEY .
				           '&userIp=' . $_SERVER['REMOTE_ADDR'] .
				           '&part=snippet,player' .
				           '&fields=items(id,player)';
				$data = json_decode( file_get_contents( $request ) );
				// get first item in returned data (we are always looking for a single video here)
				$data = $data->items[0];
				break;
			case 'vimeo':
				$request = 'https://vimeo.com/api/oembed.json?url=https%3A//vimeo.com/' . $id;
				$data = json_decode( file_get_contents( $request ) );
				break;
		}
		//error_log(__METHOD__ . ' -- ' . print_r($data, 1));
		return $data;
	}

	/**
	 * Fetch an RSS feed via shortcode
	 */
	public static function bt_rss_feed( $attrs ) {
		// @todo verify this data
		$url = $attrs['url'];
		$max = $attrs['max'];

		// Get a SimplePie feed object from the specified feed source.
		include_once( ABSPATH . WPINC . '/feed.php' );
		$rss = fetch_feed( $url );

		if ( ! is_wp_error( $rss ) ) {
			$html = '';

			// Figure out how many total items there are, but limit it
			$count = $rss->get_item_quantity( $max );

			// Build an array of all the items, starting with element 0 (first element).
			$rss_items = $rss->get_items( 0, $count );

			if ( $count == 0 ) {
				$html .= '<p class="no-items">Sorry, there are no items in the requested feed.</p>';
			} else {
			    $html .= '
                    <div class="feed">
			    ';
				foreach ( $rss_items as $item ) {
                    $html .= '
                        <div class="feed-item">
                            <a href="' . esc_url( $item->get_permalink() ) . '" target="_blank">
                                <h4>' . $item->get_title() . '</h4>
                            </a>
                            <p class="date">
			                    ' . $item->get_date( 'M' ) . '
                                <span class="day">' . $item->get_date( 'd' ) . '</span>
                            </p>
                            <p class="clearfix">' . $item->get_description() . '</p>
                        </div>
                    ';
				}
			    $html .= '
                    </div>
			    ';
			}

			return $html;
		}
	}
}
